"""Sensor class module.

This module contains the `Sensor` class, which can be used to read out the
BME280 sensor. The class provides a simple and intuitive interface, with
the main functions being `get_temperature`, `get_pressure`, and `get_humidity`.
Note that you can also get all this data at once with `get_data`, which returns
a dictionary with the three values.

For more information about how to use the `Sensor` class, have a look at the
documentation of the `Sensor` class itself.
"""

import smbus

from bme280pi.physics import calculate_abs_humidity, convert_pressure, \
    convert_temperature, round_to_n_significant_digits, pressure_at_sea_level
from bme280pi.raspberry_pi_version import detect_raspberry_pi_version
from bme280pi.readout import read_sensor


class I2CException(Exception):
    """Exception related to I2C (Inter-Integrated Circuit).

    Exception that is raised when an issue is encountered in the I2C
    section of the code. This is due to the I2C interface not being
    properly configured.
    """
    pass


class Sensor:
    """Sensor class.

    This class initializes the BME280 sensor, and offers an intuitive
    interface to access the sensor information. You can read out the
    temperature, humidity, and pressure using the corresponding get_
    commands:
        - `get_temperature`: get the current temperature
        - `get_pressure`: get the current pressure
        - `get_humidity`: get the current humidity

    The functions each support the `unit` argument, so you can specify the
    unit you would like the value to be in. For instance, the `get_temperature`
    function supports degrees C, F, or K.

    You can also use `print_data()` for a nicer presentation.

    Example usage:
    >>> sensor = Sensor()
    >>> sensor.get_temperature(unit='C')
    >>> sensor.get_temperature(unit='K')
    >>> sensor.get_humidity(relative=True)
    >>> sensor.get_humidity(relative=True)
    >>> sensor.get_pressure(unit='hPa')
    >>> sensor.get_pressure(unit='mmHg')
    """
    def __init__(self, address=0x76):
        """Initialize the sensor class.

        Carries out the following steps to initialize the class:
        - detect the Raspberry Pi version
        - initialize the bus
        - store information about the sensor in the class

        Args:
            address (int): the address of the sensor. default: 0x76
        """
        self.address = address
        self.bus = self._initialize_bus()

        self.chip_id, self.chip_version = self._get_info_about_sensor()

    def get_temperature(self, unit='C'):
        """Get a temperature reading.

        Fetches a temperature reading from the sensor and returns it in the
        user-specified temperature unit (degrees Celsius by default).
        The value can be returned in degrees Celsius (`unit='C'`),
        in degrees Fahrenheit (`unit='F'`), or in Kelvin (`unit='K'`).

        Args:
            unit (str): the unit the temperature should be in (C/F/K)

        Returns:
            float: the current temperature in the specified unit
        """
        data = self.get_data()
        return convert_temperature(data['temperature'], unit=unit)

    def get_humidity(self, relative=True):
        """Get a humidity reading.

        Fetches a humidity reading from the sensor. The value can be the
        relative humidity (`relative=True`), with values between 0 and 100.
        Alternatively, it returns the absolute humidity in kg / m^3.

        Args:
            relative (bool): indicate relative (instead of absolute) humidity

        Returns:
            float: the current relative/absolute humidity
        """
        data = self.get_data()

        if relative:
            return data['humidity']

        return calculate_abs_humidity(pressure=data['pressure'],
                                      temperature=data['temperature'],
                                      rel_humidity=data['humidity'])

    def get_pressure(self, unit='hPa', height_above_sea_level=None,
                     as_pressure_at_sea_level=False):
        """Get a pressure reading.

        Fetch the pressure from the sensor.
        The value can be returned in hPa (`unit='hPa'), Pa (`unit='Pa'),
        kPa (`unit='kPa'`), atm (`unit='atm'`), or mm Hg (`unit='mmHg'`).

        For meteorological applications, it can be useful to conver the
        pressure into its equivalent value at sea level. If you prefer this
        convention, set `as_pressure_at_sea_level=True`.
        In that case, you need to specify the height above sea level as well.

        Args:
            unit (str): unit the pressure should be in (Pa/hPa/kPa/atm/mmHg)
            height_above_sea_level (int/float): height above sea level in m
            as_pressure_at_sea_level (bool): quote as pressure above sea level

        Returns:
            float: the pressure in the specified unit, at sea level if desired
        """
        data = self.get_data()

        if as_pressure_at_sea_level:
            if height_above_sea_level is None:
                raise ValueError("You need to indicate the height above sea " +
                                 "level to get the equivalent value at sea " +
                                 "level.")
            data['pressure'] = pressure_at_sea_level(data['pressure'],
                                                     data['temperature'],
                                                     height_above_sea_level)
        return convert_pressure(data['pressure'], unit=unit)

    def get_data(self):
        """Get a reading from the sensor.

        Fetches the latest humidity, temperature, and pressure data
        from the sensor. The data is returned as a dictionary with
        keys "temperature", "humidity", and "pressure".

        Returns:
            dict: dictionary with current temperature, humidity, and pressure
        """
        return read_sensor(bus=self.bus,
                           address=self.address)

    def print_data(self, temp_unit='C', relative_humidity=True,
                   pressure_unit='hPa', n_significant_digits=4):
        """Print sensor data.

        Prints the temperature, humidity, and pressure in a easy readable
        format. The user can specify the temperature unit (e.g. "C") via
        `temp_unit`, the pressure unit (e.g. "hPa") via `pressure_unit`,
        and whether to use absolute or relative humidity via
        `relative_humidity`.

        Args:
            temp_unit (str): the unit the temperature should be in (C/F/K)
            relative_humidity (bool): relative instead of absolute humidity
            pressure_unit (str): pressure unit (Pa/hPa/kPa/atm/mmHg)
            n_significant_digits (int): number of significant digits for values

        Returns:
            None: values are printed, not returned.
        """
        data = self.get_data()
        temperature = convert_temperature(data['temperature'], temp_unit)
        pressure = convert_pressure(data['pressure'], pressure_unit)
        humidity = data['humidity']
        humidity_unit = '%'
        if not relative_humidity:
            humidity = calculate_abs_humidity(pressure=data['pressure'],
                                              temperature=data['temperature'],
                                              rel_humidity=data['humidity'])
            humidity_unit = "kg / m^3"

        # round to n significant digits
        temperature = round_to_n_significant_digits(temperature,
                                                    n_significant_digits)
        humidity = round_to_n_significant_digits(humidity,
                                                 n_significant_digits)
        pressure = round_to_n_significant_digits(pressure,
                                                 n_significant_digits)
        print("Temperature: ", temperature, temp_unit)
        print("Humidity:    ", humidity, humidity_unit)
        print("Pressure:    ", pressure, pressure_unit)

    @staticmethod
    def _initialize_bus():
        """Initialize the bus.

        Detects the raspberry pi version and initializes the bus.
        Note that the Raspberry Pi version detection is necessary because
        the first revisions needs to be initialized slightly differently.

        Returns:
            object: the bus to read data from
        """
        bus = None
        argument = 1
        if detect_raspberry_pi_version() in ['Model B R1',
                                             'Model A',
                                             'Model B+',
                                             'Model A+']:
            argument = 0

        try:
            bus = smbus.SMBus(argument)
        except FileNotFoundError:
            raise I2CException("SMBus raised a FileNotFoundError; this is " +
                               "usually due to the i2c interface being " +
                               "unconfigured. Please run 'sudo raspi-config'" +
                               " and select Interfacing Options -> I2C, " +
                               "choose and hit Enter, and then reboot. I2C " +
                               "should then be configured, and you should " +
                               "no longer see this exception")

        return bus

    def _get_info_about_sensor(self):
        """Fetch information about the sensor.

        Obtains the chip ID and version from sensor.

        Returns:
            tuple: chip ID and version as one string each
        """
        reg_id = 0xD0
        chip_id, chip_version = self.bus.read_i2c_block_data(self.address,
                                                             reg_id,
                                                             2)
        return chip_id, chip_version



from unittest import TestCase, mock

from bme280pi.raspberry_pi_version import (detect_raspberry_pi_version,
                                           get_list_of_revisions)


def raise_exception(*args, **kwargs):
    raise FileNotFoundError


class TestDetectRaspberryPiVersion(TestCase):
    def test(self):
        known_revisions = get_list_of_revisions()

        known_revisions['bad_id'] = "Unknown"

        for revision in known_revisions:
            mopen = mock.mock_open(read_data="\nRevision:" + revision + "\n")
            with mock.patch('builtins.open', mopen):
                self.assertEqual(detect_raspberry_pi_version(),
                                 known_revisions[revision])

    def test_exception(self):
        mopen = raise_exception
        with mock.patch('builtins.open', mopen):
            with self.assertWarns(Warning):
                self.assertEqual(detect_raspberry_pi_version(), "Unknown")



from unittest import TestCase

from bme280pi.readout import (get_short, get_unsigned_short, get_character,
                              get_unsigned_character, read_raw_sensor,
                              get_modified, process_calibration_data,
                              extract_raw_values,
                              improve_temperature_measurement,
                              improve_pressure_measurement,
                              improve_humidity_measurement, extract_values,
                              read_sensor)

from .sensor import FakeDataBus


class FakeRecordingBus:
    """
    This is a fake bus class (to simulate SMBus for tests)
    This "bus" class records all commands, which allows tests
    to check that the right sequence is called.
    """
    def __init__(self):
        """
        Initialize a fake bus class
        """
        self.commands = []

    def write_byte_data(self, address, register, value, force=None):
        self.commands.append(['write_byte_data', address, register,
                              value, force])

    def read_i2c_block_data(self, address, register, length, force=None):
        self.commands.append(['read_i2c_block_data', address, register,
                              length, force])
        return len(self.commands)


class TestGetShort(TestCase):
    def test(self):
        test_result = get_short(data=[129, 1, 0, 16, 44, 3, 30], index=0)
        self.assertLess(abs(test_result - 385), 1e-6)


class TestGetUnsignedshort(TestCase):
    def test(self):
        test_result = get_unsigned_short(data=[129, 1, 0, 16, 44, 3, 30],
                                         index=0)
        self.assertLess(abs(test_result - 385), 1e-6)


class TestGetCharacter(TestCase):
    def test(self):
        test_result = get_character(data=[129, 1, 0, 16, 44, 3, 30], index=0)
        self.assertLess(abs(test_result + 127), 1e-6)


class TestGetUnsignedCharacter(TestCase):
    def test(self):
        test_result = get_unsigned_character(data=[129, 1, 0, 16, 44, 3, 30],
                                             index=0)
        self.assertLess(abs(test_result - 129), 1e-6)


class TestReadRawSensor(TestCase):
    def test(self):
        bus = FakeRecordingBus()
        cals, data = read_raw_sensor(bus=bus,
                                     address="address",
                                     oversampling={'temperature': 2,
                                                   'pressure': 2,
                                                   'humidity': 2},
                                     reg_data="reg_data")
        self.assertEqual(cals[0], 3)
        self.assertEqual(cals[1], 4)
        self.assertEqual(cals[2], 5)
        self.assertEqual(data, 6)

        correct_commands = [['write_byte_data', 'address', 242, 2, None],
                            ['write_byte_data', 'address', 244, 73, None],
                            ['read_i2c_block_data', 'address', 136, 24, None],
                            ['read_i2c_block_data', 'address', 161, 1, None],
                            ['read_i2c_block_data', 'address', 225, 7, None],
                            ['read_i2c_block_data', 'address', 'reg_data', 8,
                             None]]

        for i, correct_value in enumerate(correct_commands):
            # check type of command
            self.assertEqual(bus.commands[i][0], correct_value[0])
            # check address
            self.assertEqual(bus.commands[i][1], correct_value[1])
            # check register
            self.assertEqual(bus.commands[i][2], correct_value[2])
            # check length or value
            self.assertEqual(bus.commands[i][3], correct_value[3])
            # check whether force was set
            self.assertEqual(bus.commands[i][4], correct_value[4])


def get_reference_calibration_data():
    return ([96, 110, 203, 104, 50, 0, 29, 145, 59, 215, 208, 11, 232, 38,
             42, 255, 249, 255, 172, 38, 10, 216, 189, 16],
            [75],
            [129, 1, 0, 16, 44, 3, 30])


class TestGetModified(TestCase):
    def test(self):
        cals = get_reference_calibration_data()

        self.assertEqual(get_modified(cals, 3, get_character), 268)
        self.assertEqual(get_modified(cals, 3, get_unsigned_character), 268)
        self.assertEqual(get_modified(cals, 5, get_unsigned_character,
                                      shift=True),
                         50)


class TestProcessCalibration(TestCase):
    def test(self):
        cals = get_reference_calibration_data()

        correct_dig_t = [28256, 26827, 50]
        correct_dig_p = [37149, -10437, 3024, 9960, -214, -7, 9900, -10230,
                         4285]
        correct_dig_h = [75, 385, 0, 268, 50, 30]

        dig_t, dig_p, dig_h = process_calibration_data(cals)

        # check temperature readings
        for i, dig_t_i in enumerate(dig_t):
            self.assertLess(abs(correct_dig_t[i] - dig_t_i), 1e-4)

        # check pressure readings
        for i, dig_p_i in enumerate(dig_p):
            self.assertLess(abs(correct_dig_p[i] - dig_p_i), 1e-4)

        # check humidity readings
        for i, dig_h_i in enumerate(dig_h):
            self.assertLess(abs(correct_dig_h[i] - dig_h_i), 1e-4)


class TestExtractRawValues(TestCase):
    def test(self):
        data = [76, 60, 0, 129, 49, 128, 94, 110]
        raw_pressure, raw_temperature, raw_humidity = extract_raw_values(data)
        self.assertEqual(raw_pressure, 312256)
        self.assertEqual(raw_temperature, 529176)
        self.assertEqual(raw_humidity, 24174)


class TestImproveTemperatureMeasurements(TestCase):
    def test(self):
        temp_raw = 529168
        dig_t = [28256, 26827, 50]
        temperature, t_fine = improve_temperature_measurement(temp_raw, dig_t)
        self.assertLess(abs(temperature - 2465), 1e-4)
        self.assertEqual(t_fine, 126213)


class TestImprovePressureMeasurements(TestCase):
    def test(self):
        # normal case
        raw_pressure = 312272
        dig_p = [37149, -10437, 3024, 9960, -214, -7, 9900, -10230, 4285]
        t_fine = 126240
        pressure = improve_pressure_measurement(raw_pressure, dig_p, t_fine)
        self.assertLess(abs(pressure - 96909.62784910419), 1e-4)

    def test_special_case(self):
        # special case (var1 = 0)
        raw_pressure = 312272
        dig_p = [0, 6034, -2009, -3141, 8460, 523, -7938, 6421, -4430]
        t_fine = -146.81627832355957
        pressure = improve_pressure_measurement(raw_pressure, dig_p, t_fine)
        self.assertEqual(pressure, 0)


class TestHumidityMeasurements(TestCase):
    def test(self):
        raw_humidity = 24185
        dig_h = [75, 385, 0, 268, 50, 30]
        t_fine = 126200
        humidity = improve_humidity_measurement(raw_humidity, dig_h, t_fine)
        self.assertLess(abs(humidity - 41.07923074200165), 1e-4)


class TestExtractValues(TestCase):
    def test(self):
        data = [76, 59, 0, 129, 48, 0, 94, 121]
        dig_t = [28256, 26827, 50]
        dig_p = [37149, -10437, 3024, 9960, -214, -7, 9900, -10230, 4285]
        dig_h = [75, 385, 0, 268, 50, 30]
        temp, pres, hum = extract_values(data, dig_t, dig_p, dig_h)

        self.assertLess(abs(temp - 2465.0), 1e-4)
        self.assertLess(abs(pres - 96913.34719558517), 1e-4)
        self.assertLess(abs(hum - 41.07923395171727), 1e-4)


class TestReadSensor(TestCase):
    def test(self):
        fake_data_bus = FakeDataBus(starting_point=0)
        correct_result = {'temperature': 24.65,
                          'pressure': 969.1056565652227,
                          'humidity': 41.07329061361983}

        test_result = read_sensor(bus=fake_data_bus,
                                  address="fake_address")

        for k in test_result:
            self.assertLess(abs(test_result[k] - correct_result[k]),
                            1e-4)

    def test_bad_inputs(self):
        fake_data_bus = FakeDataBus(0)
        with self.assertRaises(TypeError):
            read_sensor(bus=fake_data_bus,
                        address="fake_address",
                        oversampling="some_string")
        with self.assertRaises(TypeError):
            read_sensor(bus=fake_data_bus,
                        address="fake_address",
                        oversampling=2)
        with self.assertRaises(KeyError):
            read_sensor(bus=fake_data_bus,
                        address="fake_address",
                        oversampling={'something': 2})



"""Functions to read out and interpret the raw sensor data.

This module contains all functions necessary for reading out the raw data from
the BME280 sensor according to the Bosch data sheet, and processing this raw
data in order to obtain a more digestible format, i.e. the temperature,
pressure, and relative humidity.

The best way to use this module is through the `Sensor` class in `sensor.py`,
i.e. by initializing a `Sensor` object and using it to fetch data. One could
use the functions in this module to do the individual steps, but it's much
more practical to use the `Sensor` class directly and have it take care of
everything.

Example usage:
>>> sensor = Sensor()
>>> sensor.get_temperature(unit='C')

Have a look at the `Sensor` class in `sensor.py` for more detailed information.

The (helper) functions contained in this module are the following:
- get_short(data, index):
- get_unsigned_short(data, index):
- get_character(data, index):
- get_unsigned_character(data, index):
- read_raw_sensor(bus, address, oversampling, reg_data):
- get_modified(cal, i, function, shift=False):
- process_calibration_data(cal):
- shift_read(values, i, j, k):
- extract_raw_values(data):
- improve_temperature_measurement(temp_raw, dig_t):
- improve_pressure_measurement(raw_pressure, dig_p, t_fine):
- improve_humidity_measurement(raw_humidity, dig_h, t_fine):
- extract_values(data, dig_t, dig_p, dig_h):
- validate_oversampling(oversampling=None):
- read_sensor(bus, address, reg_data=0xF7,

Notes:
1) This module is based on the bme280 script from MattHawkinsUK,
https://bitbucket.org/MattHawkinsUK/rpispy-misc/raw/master/python/bme280.py

2) Before changing parameters in the code, it is recommended to have a look at
the data sheet available directly from Bosch:
https://www.bosch-sensortec.com/products/environmental-sensors/humidity-sensors-bme280/
"""

import time
from ctypes import c_short


def get_short(data, index):
    """Return two bytes from data as a signed 16-bit value.

    Args:
        data (list): raw data from sensor
        index (int): index entry from which to read data

    Returns:
        c_short: extracted signed 16-bit value
    """
    return c_short((data[index + 1] << 8) + data[index]).value


def get_unsigned_short(data, index):
    """Return two bytes from data as an unsigned 16-bit value.

    Args:
        data (list): raw data from sensor
        index (int): index entry from which to read data

    Returns:
        int: extracted unsigned 16-bit value
    """
    return (data[index + 1] << 8) + data[index]


def get_character(data, index):
    """Return one byte from data as a signed char.

    Args:
        data (list): raw data from sensor
        index (int): index entry from which to read data

    Returns:
        int: extracted signed char value
    """
    result = data[index]
    if result > 127:
        result -= 256
    return result


def get_unsigned_character(data, index):
    """Return one byte from data as an unsigned char.

    Args:
        data (list): raw data from sensor
        index (int): index entry from which to read data

    Returns:
        int: extracted unsigned 16-bit value
    """
    result = data[index] & 0xFF
    return result


def read_raw_sensor(bus, address, oversampling, reg_data):
    """Read raw sensor data.

    For an explanation of the parameter storage, naming, and
    data type, see Table 16, page 25, of the data sheet.
    For information about oversampling, see e.g. page 26.
    For a memory map, see Table 18 on page 27.

    Args:
        bus (object): bus from which to read data
        address (int): address at which to read data
        oversampling (dict): over-sampling rates (see data sheet)
        reg_data (int): register at which to obtain data

    Returns:
        tuple: calibration data as a list, and sensor data
    """
    control_register_address = 0xF4
    control_register_address_humidity = 0xF2
    bus.write_byte_data(address, control_register_address_humidity,
                        oversampling['humidity'])

    control1 = oversampling['temperature'] << 5
    control = control1 | oversampling['pressure'] << 2 | 1
    bus.write_byte_data(address, control_register_address, control)

    # Read blocks of calibration data from EEPROM
    # See Page 22 data sheet
    cal1 = bus.read_i2c_block_data(address, 0x88, 24)
    cal2 = bus.read_i2c_block_data(address, 0xA1, 1)
    cal3 = bus.read_i2c_block_data(address, 0xE1, 7)

    # Wait in ms
    # source: Datasheet Appendix B
    wait_time = 2.4 + 2.3 * sum(oversampling.values())

    # Wait the required time
    time.sleep(wait_time / 1000)

    # Read temperature/pressure/humidity
    data = bus.read_i2c_block_data(address, reg_data, 8)

    return (cal1, cal2, cal3), data


def get_modified(cal, i, function, shift=False):
    """Obtain and modify data from block.

    Extracts information from block, and shifts it
    if necessary.

    Args:
        cal (list): calibration data
        i (int): index from which to read
        function (object): function to apply to data
        shift (bool): whether to apply a shift to result

    Returns:
        int: processed data from specified blocks
    """
    dig = get_character(cal[2], i)
    dig = (dig << 24) >> 20
    if shift:
        return dig | (function(cal[2], 4) >> 4 & 0x0F)
    return dig | (function(cal[2], 4) & 0x0F)


def process_calibration_data(cal):
    """Process calibration data.

    Processes calibration data to extract the information pertaining
    to temperature, pressure, and humidity. Returns the relevant block
    data.

    Args:
        cal (list): calibration data

    Returns:
        tuple: block values for temperature, pressure, and humidity
    """
    dig_t = [get_unsigned_short(cal[0], 0),
             get_short(cal[0], 2),
             get_short(cal[0], 4)]

    dig_p = []
    for i in range(6, 23, 2):
        value = None
        if i == 6:
            value = get_unsigned_short(cal[0], i)
        else:
            value = get_short(cal[0], i)
        dig_p.append(value)

    dig_h = [get_unsigned_character(cal[1], 0),
             get_short(cal[2], 0),
             get_unsigned_character(cal[2], 2),
             get_modified(cal, 3, get_character),
             get_modified(cal, 5, get_unsigned_character, shift=True),
             get_character(cal[2], 6)
             ]

    return dig_t, dig_p, dig_h


def shift_read(values, i, j, k):
    """Read raw data from array and shift it.

    Reads values from array and shifts them the following way:
    - the first one is shifted to the left by 12 places
    - the second value is shifted to the left by 4 places
    - the third one is shifted to the right by 4 places
    Then a bitwise "or" operation is performed.
    Example: values = [1, 2, 3]
    - the first entry (1) is shifted 12 places to the left:
          000000000001 becomes 1000000000000
    - the second entry (2) is shifted 4 places to the left:
          000000000010 becomes 000000100000
    - the third entry (3) is shifted 4 places to the right:
          000000000011 becomes 000000000000
    Finally, a bit-wise "or" is performed:
          1000000000000 or
          0000000100000 or
          0000000000000 is
          1000000100000
    which, in decimal, is 4128.

    Args:
        values (list): the raw values
        i (int): first index to read from
        j (int): second index to read from
        k (int): third index to read from

    Returns:
        int: extracted (shifted) values
    """
    return (values[i] << 12) | (values[j] << 4) | (values[k] >> 4)


def extract_raw_values(data):
    """Extract raw reading of temperature, pressure, and humidity.

    Args:
        data (list): raw sensor data

    Returns:
        tuple: raw pressure, temperature, and humidity
    """
    raw_pressure = shift_read(data, 0, 1, 2)
    raw_temperature = shift_read(data, 3, 4, 5)
    raw_humidity = (data[6] << 8) | data[7]

    return raw_pressure, raw_temperature, raw_humidity


def improve_temperature_measurement(temp_raw, dig_t):
    """Refine the temperature measurement.

    Adapts the raw temperature measurement according to a formula specified
    in the Bosch data sheet.

    Args:
        temp_raw (int): raw temperature reading
        dig_t (list): blocks of data pertaining to temperature

    Returns:
        tuple: refined temperature measurement and reference point

    Reference:
        Bosch data sheet, Appendix A, "BME280_compensate_T_double"
    """
    var1 = ((((temp_raw >> 3) - (dig_t[0] << 1))) * (dig_t[1])) >> 11
    var2 = (((temp_raw >> 4) - (dig_t[0])) * ((temp_raw >> 4) - (dig_t[0])))
    var3 = ((var2 >> 12) * (dig_t[2])) >> 14
    t_fine = var1 + var3
    temperature = float(((t_fine * 5) + 128) >> 8)
    return temperature, t_fine


def improve_pressure_measurement(raw_pressure, dig_p, t_fine):
    """Refine the pressure measurement.

    Adapts the pressure measurement according to the formula specified in
    the Bosch data sheet, and adjust it for the available temperature
    measurement, along with the pressure readout details.

    Args:
        raw_pressure (float): raw temperature measurement
        dig_p (list): data blocks pertaining to pressure
        t_fine (float): temperature measurement

    Returns:
        float: improved pressure measurement

    Reference:
    Bosch data sheet, Appendix A, "BME280_compensate_P_double"
    """
    var1 = t_fine / 2.0 - 64000.0
    var2 = var1 * var1 * dig_p[5] / 32768.0
    var2 = var2 + var1 * dig_p[4] * 2.0
    var2 = var2 / 4.0 + dig_p[3] * 65536.0
    var1 = (dig_p[2] * var1 * var1 / 524288.0 + dig_p[1] * var1) / 524288.0
    var1 = (1.0 + var1 / 32768.0) * dig_p[0]

    if var1 == 0:
        pressure = 0
    else:
        pressure = 1048576.0 - raw_pressure
        pressure = ((pressure - var2 / 4096.0) * 6250.0) / var1
        var1 = dig_p[8] * pressure * pressure / 2147483648.0
        var2 = pressure * dig_p[7] / 32768.0
        pressure = pressure + (var1 + var2 + dig_p[6]) / 16.0

    return pressure


def improve_humidity_measurement(raw_humidity, dig_h, t_fine):
    """Refine the humidity measurement.

    Adapts the humidity measurement by using the available temperature
    information, along with the humidity readout details.

    Args:
        raw_humidity (int): raw humidity
        dig_h (list): raw data blocks pertaining to humidity measurement
        t_fine (float): temperature measurement

    Returns:
        float: refined humidity measurement

    Reference:
    Bosch data sheet, Appendix A, "BME280_compensate_H_double"
    """
    base_value = t_fine - 76800.0
    term1 = raw_humidity - (dig_h[3] * 64.0 + dig_h[4] / 16384.0 * base_value)
    term2a = base_value * (1.0 + dig_h[2] / 67108864.0 * base_value)
    term2 = dig_h[1] / 65536.0 * (1.0 + dig_h[5] / 67108864.0 * term2a)
    humidity = term1 * term2
    humidity = humidity * (1.0 - dig_h[0] * humidity / 524288.0)
    humidity = max(0, min(humidity, 100))
    return humidity


def extract_values(data, dig_t, dig_p, dig_h):
    """Extract values from raw data.

    Extracts temperature, pressure, and humidity from raw data and
    correct it to provide the best measurement.

    Args:
        data (list): data blocks from sensor
        dig_t (list): data blocks pertaining to temperature measurement
        dig_p (list): data blocks pertaining to pressure measurement
        dig_h (list): data blocks pertaining to humidity measurement

    Returns:
        tuple: three floats representing temperature, pressure, and humidity
    """
    raw_pressure, raw_temperature, raw_humidity = extract_raw_values(data)

    temperature, t_fine = improve_temperature_measurement(raw_temperature,
                                                          dig_t)
    pressure = improve_pressure_measurement(raw_pressure, dig_p, t_fine)
    humidity = improve_humidity_measurement(raw_humidity, dig_h, t_fine)

    return temperature, pressure, humidity


def validate_oversampling(oversampling=None):
    """Validate the `oversampling` parameter.

    Checks whether the parameter is valid. This parameter can either be None
    (to use the default values) or it can be a dictionary containing the
    three keys "temperature", "humidity", and "pressure".

    Args:
        oversampling (dict): None or a dictionary with over-sampling values

    Returns:
        dict: oversampling values as a dictionary
    """
    if oversampling is None:
        oversampling = {'temperature': 2,
                        'pressure': 2,
                        'humidity': 2}

    if not isinstance(oversampling, dict):
        raise TypeError("oversampling must be a dictionary")

    return oversampling


def read_sensor(bus, address, reg_data=0xF7,
                oversampling=None):
    """Read measurements from sensor.

    Reads the raw information from the sensor and converts it into a
    readable format. The function returns a dictionary with the measurements,
    with the three keys "temperature",
    See the data sheet for more information, e.g. p27 for oversampling
    settings, App. B for measurement time, Sec. 4 for data readout, etc.

    If no oversampling is defined, the code defaults to the standard 2/2/2
    for temperature/humidity/pressure. If you wish to specify your own
    oversampling parameters, please pass a dictionary to `oversampling` with
    the keys 'temperature', 'pressure', and 'humidity'.

    This module is based on the bme280 script from MattHawkinsUK, which in
    turn is based on the data sheet from Bosch (see references below).

    Args:
        bus (object): the sensor bus to read from
        address (int): the address from which to read data
        reg_data (int): register at which to obtain data
        oversampling (dict): oversampling rates

    Returns:
         dict: measurements for temperature, pressure, and humidity

    References:
    https://www.bosch-sensortec.com/products/environmental-sensors/humidity-sensors-bme280/
    """
    oversampling = validate_oversampling(oversampling=oversampling)

    cal, data = read_raw_sensor(bus=bus,
                                address=address,
                                oversampling=oversampling,
                                reg_data=reg_data)

    dig_t, dig_p, dig_h = process_calibration_data(cal)

    temperature, pressure, humidity = extract_values(data, dig_t, dig_p, dig_h)

    return {'temperature': temperature / 100.0,
            'pressure': pressure / 100.0,
            'humidity': humidity}



from unittest import TestCase

from bme280pi.physics import (validate_pressure, validate_temperature,
                              validate_height_above_sea_level,
                              validate_humidity, pressure_function,
                              calculate_abs_humidity, convert_pressure,
                              convert_temperature,
                              round_to_n_significant_digits,
                              pressure_at_sea_level)


class TestValidation(TestCase):
    def test_validate_pressure(self):
        with self.assertRaises(TypeError):
            validate_pressure("string_input")
        with self.assertRaises(ValueError):
            validate_pressure(-1)
        with self.assertRaises(ValueError):
            validate_pressure(1200)

    def test_validate_temperature(self):
        with self.assertRaises(TypeError):
            validate_temperature("string_input")
        with self.assertRaises(ValueError):
            validate_temperature(-111)
        with self.assertRaises(ValueError):
            validate_temperature(111)

    def test_validate_humidity(self):
        with self.assertRaises(TypeError):
            validate_humidity("string_input")
        with self.assertRaises(ValueError):
            validate_humidity(-1)
        with self.assertRaises(ValueError):
            validate_humidity(123)

    def test_height_above_sea_level(self):
        with self.assertRaises(TypeError):
            validate_height_above_sea_level("string_input")
        with self.assertRaises(ValueError):
            validate_height_above_sea_level(-1)
        with self.assertRaises(ValueError):
            validate_height_above_sea_level(12345)


class TestPressureFunction(TestCase):
    def test(self):
        test_values = [100, 800, 850, 900, 950, 1000, 1050, 1100]
        correct_values = [1.001176,
                          1.0040355,
                          1.0041989411764707,
                          1.0043617777777778,
                          1.004524105263158,
                          1.0046860000000002,
                          1.0048475238095238,
                          1.0050087272727273]

        for i_test, test_value in enumerate(test_values):
            test_result = pressure_function(test_value)
            correct_result = correct_values[i_test]
            self.assertLess(abs(test_result - correct_result), 1e-6)

    def test_exceptions(self):
        with self.assertRaises(ValueError):
            pressure_function(0)
        with self.assertRaises(ValueError):
            pressure_function(1111)
        with self.assertRaises(TypeError):
            pressure_function('a')


class TestCalculateAbsHumidity(TestCase):
    def test(self):
        test_pressure = [100, 800, 850, 900, 950, 1000, 1050, 1100]
        test_temp = [-10, 0, 10, 15, 20, 25, 40, 60]
        test_rel = [80, 10, 20, 90, 40, 50, 43.25, 12.34]

        correct_values = [0.0018930155819513275,
                          0.00048681001461818693,
                          0.0018843546921809028,
                          0.011566932891098143,
                          0.006927846890531939,
                          0.011536889704637605,
                          0.022155411985739303,
                          0.01612715205125449]

        for i_test, correct_value in enumerate(correct_values):
            result = calculate_abs_humidity(pressure=test_pressure[i_test],
                                            temperature=test_temp[i_test],
                                            rel_humidity=test_rel[i_test])
            self.assertLess(abs(result - correct_value), 1e-6)

    def test_valid_range(self):
        with self.assertRaises(ValueError):
            # pressure too low
            calculate_abs_humidity(pressure=0,
                                   temperature=25,
                                   rel_humidity=50)
        with self.assertRaises(ValueError):
            # pressure too high
            calculate_abs_humidity(pressure=1111,
                                   temperature=25,
                                   rel_humidity=50)
        with self.assertRaises(ValueError):
            # rel humidity too low
            calculate_abs_humidity(pressure=975,
                                   temperature=25,
                                   rel_humidity=-1)
        with self.assertRaises(ValueError):
            # rel humidity too high
            calculate_abs_humidity(pressure=975,
                                   temperature=25,
                                   rel_humidity=110)
        with self.assertRaises(ValueError):
            # temp too low
            calculate_abs_humidity(pressure=975,
                                   temperature=-101,
                                   rel_humidity=50)
        with self.assertRaises(ValueError):
            # temp too high
            calculate_abs_humidity(pressure=975,
                                   temperature=101,
                                   rel_humidity=50)

    def test_bad_arguments(self):
        with self.assertRaises(TypeError):
            calculate_abs_humidity(pressure='a',
                                   temperature=25,
                                   rel_humidity=110)
        with self.assertRaises(TypeError):
            calculate_abs_humidity(pressure=975,
                                   temperature='a',
                                   rel_humidity=110)
        with self.assertRaises(TypeError):
            calculate_abs_humidity(pressure=975,
                                   temperature=25,
                                   rel_humidity='a')


class TestConvertPressure(TestCase):
    def test(self):
        test_values = [100, 500, 850, 900, 950, 1000, 1050, 1099]

        correct_values = {'hPa': test_values,
                          'Pa': [100 * p for p in test_values],
                          'kPa': [0.1 * p for p in test_values],
                          'atm': [9.8692316931427E-4 * p for p in test_values],
                          'mmHg': [0.750062 * p for p in test_values]}

        for unit in correct_values:
            for i in range(len(correct_values[unit])):
                self.assertLess(abs(correct_values[unit][i] -
                                    convert_pressure(test_values[i],
                                                     unit=unit)),
                                1e-6)

    def test_exceptions(self):
        with self.assertRaises(ValueError):
            convert_pressure(-123)
        with self.assertRaises(ValueError):
            convert_pressure(1234)

        with self.assertRaises(TypeError):
            convert_pressure('a')
        with self.assertRaises(Exception):
            convert_pressure(123, 'UnknownUnit')


class TestConvertTemperature(TestCase):
    def test(self):
        test_values = [-50, -40, 30, 20, 10, 0, 12.34, 20, 30, 40, 50, 99]

        correct_values = {'C': test_values,
                          'F': [32 + (9. / 5) * t for t in test_values],
                          'K': [273.15 + t for t in test_values]}

        for unit in correct_values:
            for i in range(len(correct_values[unit])):
                self.assertLess(abs(correct_values[unit][i] -
                                    convert_temperature(test_values[i],
                                                        unit=unit)),
                                1e-6)

    def test_exceptions(self):
        with self.assertRaises(ValueError):
            convert_temperature(-123)
        with self.assertRaises(ValueError):
            convert_temperature(123)

        with self.assertRaises(TypeError):
            convert_temperature('a')
        with self.assertRaises(Exception):
            convert_temperature(23, 'UnknownUnit')


class TestRoundToNSignificantDigits(TestCase):
    def test(self):
        multipliers = [0.001, 0.01, 0.1, 1, 10, 100, 1000]
        test_values = [1.23456789 * m for m in multipliers]

        correct_results = {1: [1 * m for m in multipliers],
                           2: [1.2 * m for m in multipliers],
                           3: [1.23 * m for m in multipliers],
                           4: [1.235 * m for m in multipliers],
                           5: [1.2346 * m for m in multipliers]}

        for n_digits in correct_results:
            for i, test_value in enumerate(test_values):
                test_result = round_to_n_significant_digits(test_value,
                                                            n_digits)
                self.assertLess(abs(correct_results[n_digits][i] -
                                    test_result),
                                1e-6)

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            round_to_n_significant_digits('a', 2)
        with self.assertRaises(TypeError):
            round_to_n_significant_digits(2, 'a')
        with self.assertRaises(ValueError):
            round_to_n_significant_digits(2, 0)
        with self.assertRaises(ValueError):
            round_to_n_significant_digits(2, -1)
        with self.assertRaises(TypeError):
            round_to_n_significant_digits(2, 1.234)


class TestPressureAtSeaLevel(TestCase):
    def test(self):
        p = pressure_at_sea_level(pressure=972,
                                  temperature=25,
                                  height_above_sea_level=440)
        self.assertLess(abs(p - 1022.03), 0.01)

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            pressure_at_sea_level(pressure="hello",
                                  temperature=25,
                                  height_above_sea_level=440)
        with self.assertRaises(TypeError):
            pressure_at_sea_level(pressure=972,
                                  temperature="string",
                                  height_above_sea_level=440)
        with self.assertRaises(TypeError):
            pressure_at_sea_level(pressure=972,
                                  temperature=25,
                                  height_above_sea_level="string")
        with self.assertRaises(ValueError):
            pressure_at_sea_level(pressure=1234,
                                  temperature=25,
                                  height_above_sea_level=440)
        with self.assertRaises(ValueError):
            pressure_at_sea_level(pressure=972,
                                  temperature=123,
                                  height_above_sea_level=440)
        with self.assertRaises(ValueError):
            pressure_at_sea_level(pressure=972,
                                  temperature=25,
                                  height_above_sea_level=11111)



"""Detect the Raspberry Pi Version.

Contains the detect_raspberry_pi_version function, which
detects the version of the Raspberry Pi used.
"""
import warnings


def get_list_of_revisions():
    """List of known Raspberry Pi Revisions.

    Provides a list of known Raspberry Pi CPU IDs and the corresponding
    Raspberry Pi model name ("revision").

    Args:

    Returns:
        dict: dictionary of Raspberry Pi Revisions
    """
    known_revisions = {'0002': 'Model B R1',
                       '0003': 'Model B R1',
                       '0004': 'Model B R2',
                       '0005': 'Model B R2',
                       '0006': 'Model B R2',
                       '0007': 'Model A',
                       '0008': 'Model A',
                       '0009': 'Model A',
                       '000d': 'Model B R2',
                       '000e': 'Model B R2',
                       '000f': 'Model B R2',
                       '0010': 'Model B+',
                       '0011': 'Compute Module',
                       '0012': 'Model A+',
                       'a01041': 'Pi 2 Model B',
                       'a21041': 'Pi 2 Model B',
                       '900092': 'Pi Zero',
                       '900093': 'Pi Zero',
                       'a02082': 'Pi 3 Model B',
                       'a22082': 'Pi 3 Model B',
                       '9000c1': 'Pi Zero W',
                       'c03111': 'Pi 4 Model B',
                       'abcdef': 'TestModel',
                       '0000': 'Unknown'}
    return known_revisions


def detect_raspberry_pi_version():
    """Detect the Raspberry Pi Version.

    Detects the Raspberry Pi version based on CPU information.
    Note that if the model comes back as "Unknown", you may need
    to add it to the dictionary in `get_list_of_revisions`. The
    current list is up-to-date as of June 2020.

    Args:

    Returns:
        str: Raspberry Pi version
    """
    revision = "0000"
    known_revisions = get_list_of_revisions()
    try:
        cpu_info = open("/proc/cpuinfo", "r")
        for line in cpu_info:
            if line.find("Revision") >= 0:
                revision = line.rstrip().replace(" ", "").split(":")[1]
        cpu_info.close()
    except FileNotFoundError:
        warnings.warn("Could not find /proc/cpuinfo")
        return "Unknown"

    if revision in known_revisions:
        return known_revisions[revision]

    return "Unknown"



"""Physics functions for pressure/temperature/humidity/...

Provides functions related to converting humidity from relative
to absolute humidity. Also provides functions for converting
pressure & temperature into different units (e.g. Kelvin, mm Hg),
and to round numbers to a n significant digits.

The following functions are present in this module:
- validate_pressure(pressure)
- validate_temperature(temperature)
- validate_humidity(rel_humidity)
- validate_height_above_sea_level(height_above_sea_level)
- pressure_function(pressure)
- calculate_abs_humidity(pressure, temperature, rel_humidity)
- convert_pressure(pressure, unit='hPa')
- convert_temperature(temperature, unit='C')
- pressure_at_sea_level(pressure, temperature, height_above_sea_level)
- round_to_n_significant_digits(value, n_digits)

source for formulae:
https://planetcalc.com/2167/
https://keisan.casio.com/keisan/image/Convertpressure.pdf
"""

import math


def validate_pressure(pressure):
    """Validate input pressure.

    Checks that pressure satisfies the following constraints:
    - needs to be positive
    - needs to be smaller than 1100 (largest value ever was 1083)

    A `ValueError` is raised if any of the assumptions are violated.

    Args:
        pressure (float/int): pressure in hPa

    Returns:
        None
    """
    if not isinstance(pressure, (float, int)):
        raise TypeError("Pressure must be int or float")

    if pressure <= 0 or pressure > 1100:
        raise ValueError("Pressure must be between 0 and 1100")


def validate_temperature(temperature):
    """Validate input temperature.

    Checks that temperature satisfies the following constraints:
    - needs to be smaller than 100 degrees (humidity calculations
      don't make much sense above this temperature)
    - needs to be larger than -100 (same reason)

    A `ValueError` is raised if any of the assumptions are violated.

    Args:
        temperature (float/int): temperature in degrees Celsius

    Returns:
        None
    """
    if not isinstance(temperature, (float, int)):
        raise TypeError("Temperature must be int or float")

    if temperature < -100 or temperature > 100:
        raise ValueError("Temperature must be between -100 and +100")


def validate_humidity(rel_humidity):
    """Validate input humidity.

    Checks that humidity satisfies the following constraints:
    - relative humidity must be below 100%
    - relative humidity must be above 0%

    A `ValueError` is raised if any of the assumptions are violated.

    Args:
        rel_humidity (float/int): relative humidity in percent (i.e. 0-100)

    Returns:
        None
    """
    if not isinstance(rel_humidity, (float, int)):
        raise TypeError("Relative Humidity must be int or float")
    if rel_humidity < 0 or rel_humidity > 100:
        raise ValueError("Rel. humidity must be between 0 and 100")


def validate_height_above_sea_level(height_above_sea_level):
    """Validate height above sea level.

    Checks that height above sea level satisfies the following constraints:
    - needs to be positive
    - needs to be smaller than 11000 (limit of validity of conversion formula)

    A `ValueError` is raised if any of the assumptions are violated.

    Args:
        height_above_sea_level (float/int): height above sea level in meters

    Returns:
        None
    """
    if not isinstance(height_above_sea_level, (float, int)):
        raise TypeError("Height above sea level must be int or float")

    if height_above_sea_level <= 0 or height_above_sea_level > 11000:
        raise ValueError("Height above sea level must be between zero " +
                         "and 11000")


def pressure_function(pressure):
    """Saturation vapor pressure function.

    Calculates the relevant factor to convert the saturation vapor pressure
    in pure phase to the saturation vapor pressure in moist air.

    Args:
        pressure (int/float): pressure in hPa

    Returns:
        float: factor to convert saturation vapor pressure
    """
    validate_pressure(pressure)
    return 1.0016 + 3.16 * 1e-6 * pressure - 0.074 / pressure


def calculate_abs_humidity(pressure, temperature, rel_humidity):
    """Calculate the absolute humidity.

    The absolute humidity is calculated in the following steps:
    - we first calculate the saturation vapor pressure in pure phase (e_w)
    - we use the pressure function f(p) to calculate the saturation vapor
        pressure of moist air (e_w_moist)
    - We can then calculate the absolute humidity using the formulae below.

    We start with the ideal gas law, PV = m R T,
        PV = (m/M) R T
    where R is the universal gas constant (8.314 kg m^2 / s^2 mol K), and
    transform it to
        eV = m R_v T
    where R_v is the specific gas constant for water vapor (461.5 J / kg K).
    Then,
        m / V = e / (R_v T)
    which is the absolute humidity (i.e. mass of water vapor per volume).

    Args:
        pressure (int/float): pressure in hPa
        temperature (int/float): temperature in degrees Celsius
        rel_humidity (int/float): relative humidity in percent (0-100)

    Returns:
        float: humidity measurement value
    """
    validate_pressure(pressure)
    validate_temperature(temperature)
    validate_humidity(rel_humidity)

    # saturation vapor pressure in pure phase
    e_w = 6.112 * math.exp(17.62 * temperature / (243.12 + temperature))

    # saturation vapor pressure in moist air
    e_w_moist = pressure_function(pressure=pressure) * e_w

    # actual vapor pressure
    vapor_pressure = e_w_moist * rel_humidity

    # absolute humidity
    return vapor_pressure / (461.5 * (273.15 + temperature))


def convert_pressure(pressure, unit='hPa'):
    """Pressure in user-specified unit.

    Converts pressure from hPa (input) to the desired unit.
    Available options are:
     - hPa (`unit='hPa'`)
     - Pa (`unit='Pa'`)
     - kPa (`unit='kPa'`)
     - atm (`unit='atm'`)
     - mm Hg (`unit='mmHg'`)

    Args:
        pressure (int/float): pressure in hPa
        unit (str): unit TO CONVERT PRESSURE TO (hPa/Pa/kPa/atm/mmHg)

    Returns:
        float: pressure in specified unit
    """
    validate_pressure(pressure)

    conversion_factor = {"hPa": 1,
                         "Pa": 100,
                         "kPa": 0.1,
                         "atm": 9.8692316931427E-4,
                         "mmHg": 0.750062}

    if unit in conversion_factor:
        return conversion_factor[unit] * pressure

    raise Exception("Unknown pressure unit: " + unit)


def convert_temperature(temperature, unit='C'):
    """Temperature in user-specified unit.

    Converts temperature from Celsius (input) to the desired unit.
    Available options are:
    - Celsius (`unit='C'`)
    - Fahrenheit (`unit='F'`)
    - Kelvin (`unit='K'`)

    Args:
        temperature (int/float): temperature in degrees Celsius
        unit (str): unit to convert the temperature to (C/F/K)

    Returns:
        float: temperature in desired unit
    """
    validate_temperature(temperature)

    if unit == 'C':
        return temperature

    if unit == 'F':
        return temperature * (9. / 5) + 32.

    if unit == 'K':
        return temperature + 273.15

    raise Exception("Unknown temperature unit: " + unit)


def pressure_at_sea_level(pressure, temperature, height_above_sea_level):
    r"""Convert pressure to pressure at sea level.

    Uses a simple formula to convert the observed pressure to the equivalent
    pressure at sea level. The pressure at sea level is a commonly quoted
    quantity, often referred to as QFF whereas the "local" observed pressure
    is referred to as QFE.

    .. math::

        e = -\frac{g}{R_d \gamma}
        f = \left(1 + \frac{\gamma \cdot h}{T - \gamma\cdot h}\right)
        p = p_0 * f ^ e

    where `e` is the exponent, `f` is the correction factor, `gamma` is the
    derivative `dT/dz` (which is approx. -0.0065 K/m), g is the gravitational
    acceleration in free fall (9.80665 m/s^2), `T` is the temperature,
    `R_d` is the specific gas constant of dry air (287 J/kg/K), `p` is the
    observed pressure, and `p_0` is the pressure at sea level. All calculations
    are in SI units.

    Args:
        pressure (float/int): pressure in hPa
        temperature (float/int): temperature in degrees Celsius
        height_above_sea_level (float/int): height above sea level in meters

    Returns:
        float: equivalent pressure at sea level in hPa
    """
    validate_pressure(pressure)
    validate_temperature(temperature)
    validate_height_above_sea_level(height_above_sea_level)

    gamma = -0.0065
    gravitational_acc = 9.80665
    r_d = 287
    temp = convert_temperature(temperature, unit='K')

    exponent = - gravitational_acc / (r_d * gamma)
    nominator = gamma * height_above_sea_level
    denominator = (temp - gamma * height_above_sea_level)
    base = 1 + nominator / denominator

    return pressure / pow(base, exponent)


def round_to_n_significant_digits(value, n_digits):
    """Round to n significant digits.

    Rounds a number to n significant digits, e.g. for 1234 the result
    with 2 significant digits would be 1200.

    Args:
        value (float/int): the value to be rounded
        n_digits (int): the desired number of significant digits

    Returns:
        float: the value rounded to the desired number of significant digits
    """
    if not isinstance(value, (float, int)):
        raise TypeError("Value must be int or float")
    if not isinstance(n_digits, int):
        raise TypeError("Number of digits must be int")
    if not n_digits > 0:
        raise ValueError("Number of digits must be greater than zero")

    return round(value, n_digits - 1 - int(math.floor(math.log10(abs(value)))))



"""bme280pi package: a package to read out the BME280 sensor on Raspberry Pi.

Here's a quick overview of how this package can be used:

1) Import the Sensor class:
`from bme280pi import Sensor`
2) Initialize a `Sensor` object:
`sensor = Sensor()`
3) Get the sensor data:
`data = sensor.get_data()`

You can also just fetch the temperature, the pressure, or the humidity with
the individual `.get_temperature()`, `.get_pressure`, and `get_humidity`
functions.

The standard units are degrees Celsius (C) for temperature, hectopascal (hPa)
for pressure, and percentage for humidity. Other units are supported for
temperature (F and K) and pressure (mm Hg, atm, Pa, kPa), and the humidity
can also be indicated in absolute terms (i.e. g per m^3).

For more in-depth information see the individual docstrings of the functions
of the `Sensor` class.
"""

from .sensor import Sensor

__all__ = ["Sensor"]



# !/usr/bin/env python3

"""bme280pi: the BME280 Sensor Reader for Raspberry Pi.

This package provides:
- an intuitive way to access the Bosch BME280 sensor
- the ability to convert measurements into different units
"""

import os
import sys

from setuptools import setup, find_packages

here = os.path.dirname(__file__)
README = open(os.path.join(here, 'README.md')).read()


if sys.version_info.major < 3 or sys.version_info.minor < 5:
    print("Package is for python 3.5 or later, please upgrade")
    sys.exit(1)

github_url = "https://github.com/MarcoAndreaBuchmann/bme280pi/"
download_url = github_url + "archive/v1.0.tar.gz"


setup(name='bme280pi',
      version='1.1.0',
      license='MIT',
      description="bme280pi: the BME280 Sensor Reader for Raspberry Pi",
      long_description=README,
      long_description_content_type="text/markdown",
      author='Marco-Andrea Buchmann',
      author_email='bmarcoa@gmail.com',
      url='https://www.github.com/MarcoAndreaBuchmann/bme280pi',
      download_url=download_url,
      project_urls={"Bug Tracker": github_url + "issues",
                    "Source Code": github_url},
      keywords=['Raspberry', 'Pi', 'Raspberry Pi', 'BME280', 'sensor',
                'readout', 'temperature', 'pressure', 'humidity'],
      platforms=["Linux"],
      python_requires=">=3.5",
      packages=find_packages(),
      tests_require=["smbus", "pytest"],
      install_requires=['smbus'],
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'Topic :: Software Development :: Build Tools',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7']
      )



import io
from unittest import TestCase, mock

from bme280pi.sensor import Sensor, I2CException


class FakeSMBus:
    def __init__(self, value):
        """
        This is a fake bus class (to replace SMBus).
        It records the value it is initialized with, and makes
        it accessible for checks. It also reports its chip ID and
        version as fake.
        """
        self.value = value

    @staticmethod
    def read_i2c_block_data(address, register, length, force=None):
        return "fake_chip_id", "fake_version"


class TestInitializeBus(TestCase):
    def test(self):
        # this test requires us to override the processor type and smbbus
        known_revisions = {'0002': 0,
                           '0003': 0,
                           '0004': 1,
                           '0005': 1,
                           '0006': 1,
                           '0007': 0,
                           '0008': 0,
                           '0009': 0,
                           '000d': 1,
                           '000e': 1,
                           '000f': 1,
                           '0010': 0,
                           '0011': 1,
                           '0012': 0,
                           'a01041': 1,
                           'a21041': 1,
                           '900092': 1,
                           '900093': 1,
                           'a02082': 1,
                           'a22082': 1,
                           '9000c1': 1,
                           'c03111': 1,
                           'abcdef': 1,
                           '0000': 1}

        for revision in known_revisions:
            m = mock.mock_open(read_data="\nRevision:" + revision + "\n")
            with mock.patch('builtins.open', m):
                with mock.patch('smbus.SMBus', FakeSMBus):
                    sensor = Sensor()
                    self.assertEqual(sensor.bus.value,
                                     known_revisions[revision])


class FakeDataBus:
    def __init__(self, starting_point=-1):
        """
        A further fake bus class (to replace SMBus).
        This version returns data that is a realistic representation
        of actual data
        """
        self.i_read = starting_point
        self.data = [['fake_chip_id', 'fake_version'],
                     [96, 110, 203, 104, 50, 0, 29, 145, 59, 215, 208, 11,
                      232, 38, 42, 255, 249, 255, 172, 38, 10, 216, 189, 16],
                     [75],
                     [129, 1, 0, 16, 44, 3, 30],
                     [76, 60, 128, 129, 49, 128, 94, 120]]

    def write_byte_data(self, address, register, value, force=None):
        pass

    def read_i2c_block_data(self, address, register, length, force=None):
        self.i_read += 1
        return self.data[self.i_read]


class FileNotFoundSMBus:
    def __init__(self, bus_address):
        """
        This fake bus class raises a FileNotFoundError, because that's what
        SMBus can do in case I2C is not configured.
        """
        raise FileNotFoundError("No such file or directory")


def initialize_fake_bus(*args, **kwargs):
    return FakeDataBus()


class TestSensor(TestCase):
    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test_get_data(self):
        sensor = Sensor()
        self.assertEqual(sensor.chip_id, "fake_chip_id")
        self.assertEqual(sensor.chip_version, "fake_version")

        data = sensor.get_data()
        self.assertLess(abs(data['temperature'] - 24.65), 1e-4)
        self.assertLess(abs(data['pressure'] - 969.1056565652227), 1e-4)
        self.assertLess(abs(data['humidity'] - 41.07329061361983), 1e-4)

    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test_get_temperature(self):
        sensor = Sensor()
        temperature = sensor.get_temperature()
        self.assertLess(abs(temperature - 24.65), 1e-4)

    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test_get_pressure(self):
        sensor = Sensor()
        pressure = sensor.get_pressure()
        self.assertLess(abs(pressure - 969.1056565652227), 1e-4)

    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test_get_pressure_above_sea_level(self):
        sensor = Sensor()
        pressure = sensor.get_pressure(height_above_sea_level=440,
                                       as_pressure_at_sea_level=True)
        self.assertLess(abs(pressure - 1019.0420210), 1e-4)

    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test_get_pressure_without_height_above_sea_level(self):
        sensor = Sensor()
        with self.assertRaises(ValueError):
            sensor.get_pressure(height_above_sea_level=None,
                                as_pressure_at_sea_level=True)

    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test_get_humidity(self):
        sensor = Sensor()
        humidity = sensor.get_humidity()
        self.assertLess(abs(humidity - 41.07329061361983), 1e-4)

        sensor = Sensor()
        humidity = sensor.get_humidity(relative=False)
        self.assertLess(abs(humidity - 0.009291279797753835), 1e-4)

    def test_unconfigured_i2c(self):
        with mock.patch('smbus.SMBus', FileNotFoundSMBus):
            with self.assertRaises(I2CException):
                Sensor()


class TestPrintSensor(TestCase):
    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test(self):
        sensor = Sensor()
        self.assertEqual(sensor.chip_id, "fake_chip_id")
        self.assertEqual(sensor.chip_version, "fake_version")

        ref_message = "Temperature:  24.65 C\n" + \
                      "Humidity:     41.07 %\n" + \
                      "Pressure:     969.1 hPa\n"

        with mock.patch("sys.stdout", new_callable=io.StringIO) as fake_out:
            sensor.print_data()
            self.assertEqual(fake_out.getvalue(), ref_message)

    @mock.patch("bme280pi.Sensor._initialize_bus", initialize_fake_bus)
    def test_print_with_absolute_humidity(self):
        sensor = Sensor()
        ref_message = "Temperature:  24.65 C\n" + \
                      "Humidity:     0.009291 kg / m^3\n" + \
                      "Pressure:     969.1 hPa\n"

        with mock.patch("sys.stdout", new_callable=io.StringIO) as fake_out:
            sensor.print_data(relative_humidity=False)
            self.assertEqual(fake_out.getvalue(), ref_message)
